// Title:	Entity.cs
// Author: 	Scott Ellington <scott.ellington@gmail.com>
//
// Copyright (C) 2006 Scott Ellington and authors
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using CsGL.OpenGL;

namespace SalmonViewer
{
	public struct TexCoord 
	{
		public float U;
		public float V;

		public TexCoord ( float u , float v )
		{
			U = u;
			V = v;
		}
	}
	
	public class Entity : IRenderable
	{
		// TODO: OO this
		// fields should be private
		// constructor with verts and faces
		// normalize in ctor
		
		int glListNo = -1;
		
		public Material material = new Material ();
		
		// The stored vertices 
		public Vector[] vertices;

		// The calculated normals
		public Vector[] normals;
		
		// The indices of the triangles which point to vertices
		public Triangle[] indices;


		public Quad[] quads;
		
		// The coordinates which map the texture onto the entity
		public TexCoord[] texcoords;

		bool normalized = false;
		
		/// <summary>
		/// Calculates the surface normals for all vertices.
		/// A surface normal is a vector perpendicular to the tangent plane to that surface.
		/// http://en.wikipedia.org/wiki/Surface_normal
		/// </summary>
		public void CalculateNormals ()
		{
			if ( indices == null ) return;
			
			// a normal is created for each vertex
			normals = new Vector [vertices.Length];

			// first let's create a surface normal for each triangle
			Vector[] temps = new Vector [ indices.Length ];
			for ( int ii=0 ; ii < indices.Length ; ii++ )
			{
				Triangle tr = indices [ii];
				
				Vector v1 = vertices [ tr.Vertex1 ] - vertices  [ tr.Vertex2 ];
				Vector v2 = vertices [ tr.Vertex2 ] - vertices  [ tr.Vertex3 ];

				temps [ii] = v1.CrossProduct ( v2 );
			}
			
			// then merge the triangle's vectors for each vertex
			for ( int ii = 0; ii < vertices.Length ; ii++ )
			{
				// create a new origin vector (0,0,0)
				Vector v = new Vector ();
				
				// we'll do this by looping through all of the triangles
				for ( int jj = 0; jj < indices.Length ; jj++ )
				{
					Triangle tr = indices [jj];
					
					// and check if this triangle shares this vertex
					if ( tr.Vertex1 == ii || tr.Vertex2 == ii || tr.Vertex3 == ii )
					{
						// if so, add this vector to our vector
						// which has the effect of combining the two's magnitude and direction
						// causing a smoothing effect on the entity
						v += temps [jj];
					}
				}

				// finally normalize the vector
				normals [ii] = v.Normalize ();
			}

			normalized = true;
		}

		public void Render ()
		{
			if ( indices == null ) return;
			
			// use a GlList for performance
			if (glListNo >= 0)
			{
				GL.glCallList((uint)glListNo);
				return;
			}
			
			// first run we create the glList and manually render our entity
			
			glListNo = (int)GL.glGenLists(1);
			GL.glNewList((uint)glListNo, GL.GL_COMPILE);

			// set up light properties
			GL.glMaterialfv (GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, material.Ambient);
			GL.glMaterialfv (GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, material.Diffuse);
			GL.glMaterialfv (GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, material.Specular);
			GL.glMaterialf (GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, material.Shininess);
	
			// if we have a texture, bind it
			if (material.TextureId >= 0 ) 
			{
				GL.glBindTexture ( GL.GL_TEXTURE_2D, (uint)material.TextureId ); 
				GL.glEnable( GL.GL_TEXTURE_2D );
			}

			// Draw every triangle in the entity, it's normal, and it's material (if exists)
			GL.glBegin ( GL.GL_TRIANGLES);		
			foreach ( Triangle tri in indices )
			{ 
				// Vertex 1
				if (normalized) GL.glNormal3d ( normals[tri.Vertex1].X, normals[tri.Vertex1].Y, normals[tri.Vertex1].Z );
				if ( material.TextureId >= 0 ) GL.glTexCoord2f ( texcoords [ tri.Vertex1 ].U, texcoords [ tri.Vertex1 ].V);
				GL.glVertex3d ( vertices[tri.Vertex1].X, vertices[tri.Vertex1].Y, vertices[tri.Vertex1].Z );

				// Vertex 2
				if (normalized) GL.glNormal3d ( normals[tri.Vertex2].X, normals[tri.Vertex2].Y, normals[tri.Vertex2].Z );
				if ( material.TextureId >= 0 ) GL.glTexCoord2f ( texcoords [ tri.Vertex2 ].U, texcoords [ tri.Vertex2 ].V);
				GL.glVertex3d ( vertices[tri.Vertex2].X, vertices[tri.Vertex2].Y, vertices[tri.Vertex2].Z );

				// Vertex 3
				if (normalized) GL.glNormal3d ( normals[tri.Vertex3].X, normals[tri.Vertex3].Y, normals[tri.Vertex3].Z );
				if ( material.TextureId >= 0 ) GL.glTexCoord2f( texcoords [ tri.Vertex3 ].U, texcoords [ tri.Vertex3 ].V);
				GL.glVertex3d ( vertices[tri.Vertex3].X, vertices[tri.Vertex3].Y, vertices[tri.Vertex3].Z );
			}
			
			GL.glEnd();
			GL.glDisable( GL.GL_TEXTURE_2D );
			
			GL.glEndList();
		}	
	}
}
